"""Test milc.questions
"""
from unittest.mock import patch

import milc.questions


@patch('builtins.input', return_value='y', autospec=True)
@patch('builtins.print', autospec=True)
def test_yesno_y(mock_inputs, mock_outputs):
    """Make sure an answer of 'y' returns True
    """
    answer = milc.questions.yesno('')
    assert answer is True


@patch('builtins.input', return_value='n', autospec=True)
@patch('builtins.print', autospec=True)
def test_yesno_n(mock_inputs, mock_outputs):
    """Make sure an answer of 'n' returns False
    """
    assert milc.questions.yesno('') is False


@patch('builtins.input', return_value='', autospec=True)
@patch('builtins.print', autospec=True)
def test_yesno_default(mock_inputs, mock_outputs):
    """Make sure a blank answer returns True
    """
    assert milc.questions.yesno('', default=True) is True


@patch('builtins.input', return_value='hi', autospec=True)
@patch('builtins.print', autospec=True)
def test_question(mock_inputs, mock_outputs):
    """Make sure we get 'hi' back from the input.
    """
    assert milc.questions.question('') == 'hi'


@patch('builtins.input', return_value='', autospec=True)
@patch('builtins.print', autospec=True)
def test_question_default(mock_inputs, mock_outputs):
    """Make sure a blank answer returns True
    """
    assert milc.questions.question('', default=True) is True


@patch('builtins.input', return_value='1', autospec=True)
@patch('builtins.print', autospec=True)
def test_choice_1(mock_inputs, mock_outputs):
    """Make we get back the right answer.
    """
    assert milc.questions.choice('', [True, False, 'hi']) is True


@patch('builtins.input', return_value='3', autospec=True)
@patch('builtins.print', autospec=True)
def test_choice_3(mock_inputs, mock_outputs):
    """Make we get back the right answer.
    """
    assert milc.questions.choice('', [True, False, 'hi']) == 'hi'


@patch('builtins.input', return_value='', autospec=True)
@patch('builtins.print', autospec=True)
def test_choice_default(mock_inputs, mock_outputs):
    """Make we get back the right answer.
    """
    assert milc.questions.choice('', [True, False, 'hi'], default=2) == 'hi'



"""Unit tests for milc.attrdict.
"""
import milc.attrdict


def attrdict_setup(attrdict=milc.attrdict.AttrDict):
    """Setup a simple AttrDict object for testing.
    """
    simple_dict = attrdict()
    simple_dict['a'] = True
    simple_dict['b'] = False

    return simple_dict


def test_AttrDict_dictionary():
    """Test AttrDict dictionary access
    """
    simple_dict = attrdict_setup()
    assert simple_dict['a'] is True
    assert simple_dict['b'] is False


def test_AttrDict_attribute():
    """Test AttrDict attribute access
    """
    simple_dict = attrdict_setup()
    assert simple_dict.a is True
    assert simple_dict.b is False


def test_AttrDict_invalid_dictionary():
    """Make sure invalid dictionary keys raise KeyError
    """
    simple_dict = attrdict_setup()

    try:
        simple_dict['c']
        raise KeyError('Should not be able to access non-existent key "c".')
    except KeyError:
        pass


def test_AttrDict_invalid_attribute():
    """Make sure invalid attributes raise KeyError
    """
    simple_dict = attrdict_setup()

    try:
        simple_dict.c
        raise KeyError('Should not be able to access non-existent attribute "c".')
    except KeyError:
        pass


def test_AttrDict_iteration():
    """Make sure we can iterate over our stored volues.
    """
    for key, value in attrdict_setup().items():
        if key == 'a':
            assert value is True
        elif key == 'b':
            assert value is False
        else:
            raise KeyError('Unexpected key found in attrdict: ' + key)


def test_SparseAttrDict_dictionary():
    """Test SparseAttrDict dictionary access.
    """
    simple_dict = attrdict_setup(milc.attrdict.SparseAttrDict)

    assert simple_dict['a'] is True
    assert simple_dict['b'] is False
    assert simple_dict['c'] is None


def test_SparseAttrDict_attribute():
    """Test SparseAttrDict attribute access.
    """
    simple_dict = attrdict_setup(milc.attrdict.SparseAttrDict)

    assert simple_dict.a is True
    assert simple_dict.b is False
    assert simple_dict.c is None



"""Tests for milc/ansi.py

TODO:
    * skullydazed/anyone: Write tests for the log format classes
"""
import milc.ansi


def test_format_ansi():
    string = ''
    for color in milc.ansi.ansi_colors:
        string = string + '{' + color + '}'
    string = milc.ansi.format_ansi(string)

    assert string == '\x1b[30m\x1b[34m\x1b[36m\x1b[32m\x1b[90m\x1b[94m\x1b[96m\x1b[92m\x1b[95m\x1b[91m\x1b[97m\x1b[93m\x1b[35m\x1b[31m\x1b[39m\x1b[37m\x1b[33m\x1b[40m\x1b[44m\x1b[46m\x1b[42m\x1b[100m\x1b[104m\x1b[106m\x1b[102m\x1b[105m\x1b[101m\x1b[107m\x1b[103m\x1b[45m\x1b[41m\x1b[49m\x1b[47m\x1b[43m\x1b[1m\x1b[2m\x1b[22m\x1b[0m\x1b[0m'



from .attrdict import AttrDict


class Configuration(AttrDict):
    """Represents the running configuration.

    This class never raises IndexError, instead it will return None if a
    section or option does not yet exist.
    """
    def __getitem__(self, key):
        """Returns a config section, creating it if it doesn't exist yet.
        """
        if key not in self._data:
            self.__dict__[key] = self._data[key] = ConfigurationSection(self)

        return self._data[key]


class ConfigurationSection(Configuration):
    def __init__(self, parent, *args, **kwargs):
        super(ConfigurationSection, self).__init__(*args, **kwargs)
        self.parent = parent

    def __getitem__(self, key):
        """Returns a config value, pulling from the `user` section as a fallback.
        This is called when the attribute is accessed either via the get method or through [ ] index.
        """
        if key in self._data and self._data.get(key) is not None:
            return self._data[key]

        elif key in self.parent.user:
            return self.parent.user[key]

        return None

    def __getattr__(self, key):
        """Returns the config value from the `user` section.
        This is called when the attribute is accessed via dot notation but does not exist.
        """
        if key in self.parent.user:
            return self.parent.user[key]

        return None


class SubparserWrapper(object):
    """Wrap subparsers so we can track what options the user passed.
    """
    def __init__(self, cli, submodule, subparser):
        self.cli = cli
        self.submodule = submodule
        self.subparser = subparser

        for attr in dir(subparser):
            if not hasattr(self, attr):
                setattr(self, attr, getattr(subparser, attr))

    def completer(self, completer):
        """Add an arpcomplete completer to this subcommand.
        """
        self.subparser.completer = completer

    def add_argument(self, *args, **kwargs):
        """Add an argument for this subcommand.

        This also stores the default for the argument in `self.cli.default_arguments`.
        """
        if kwargs.get('action') == 'store_boolean':
            # Store boolean will call us again with the enable/disable flag arguments
            return handle_store_boolean(self.cli, *args, **kwargs)

        self.cli.acquire_lock()
        argument_name = get_argument_name(self.cli, *args, **kwargs)
        self.subparser.add_argument(*args, **kwargs)

        if kwargs.get('action') == 'store_false':
            self.cli._config_store_false.append(argument_name)

        if kwargs.get('action') == 'store_true':
            self.cli._config_store_true.append(argument_name)

        if self.submodule not in self.cli.default_arguments:
            self.cli.default_arguments[self.submodule] = {}

        self.cli.default_arguments[self.submodule][argument_name] = kwargs.get('default')
        self.cli.release_lock()


def get_argument_name(self, *args, **kwargs):
    """Takes argparse arguments and returns the dest name.
    """
    try:
        return self._arg_parser._get_optional_kwargs(*args, **kwargs)['dest']
    except ValueError:
        return self._arg_parser._get_positional_kwargs(*args, **kwargs)['dest']


def handle_store_boolean(self, *args, **kwargs):
    """Does the add_argument for action='store_boolean'.
    """
    disabled_args = None
    disabled_kwargs = kwargs.copy()
    disabled_kwargs['action'] = 'store_false'
    disabled_kwargs['dest'] = get_argument_name(self, *args, **kwargs)
    disabled_kwargs['help'] = 'Disable ' + kwargs['help']
    kwargs['action'] = 'store_true'
    kwargs['help'] = 'Enable ' + kwargs['help']

    for flag in args:
        if flag[:2] == '--':
            disabled_args = ('--no-' + flag[2:],)
            break

    self.add_argument(*args, **kwargs)
    self.add_argument(*disabled_args, **disabled_kwargs)

    return (args, kwargs, disabled_args, disabled_kwargs)



from milc import cli


def check_command(command, *args):
    cmd = [command] + list(args)
    return cli.run(cmd, combined_output=True)


def check_returncode(result, expected=0):
    """Print stdout if `result.returncode` does not match `expected`.
    """
    if result.returncode != expected:
        print('`%s` stdout:' % ' '.join(result.args))
        print(result.stdout)
        print('returncode:', result.returncode)
    assert result.returncode == expected


def test_hello():
    result = check_command('./hello')
    check_returncode(result)
    assert result.stdout == '\x1b[34mℹ\x1b[0m Hello, World, from cli.log.info!\x1b[0m\nHello, World, from cli.echo!\n'


def test_hello_help():
    result = check_command('./hello', '--help')
    check_returncode(result)
    assert '[-n NAME]' in result.stdout
    assert '[--no-comma]' in result.stdout


def test_hello_no_color():
    result = check_command('./hello', '--no-color')
    check_returncode(result)
    assert result.stdout == 'ℹ Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n'


def test_hello_no_color_no_comma():
    result = check_command('./hello', '--no-color', '--no-comma')
    check_returncode(result)
    assert result.stdout == 'ℹ Hello World, from cli.log.info!\nHello World, from cli.echo!\n'


def test_hello_no_color_log_file():
    result = check_command('./hello', '--no-color', '--log-file', '/dev/stdout')
    check_returncode(result)
    assert result.stdout.startswith('[INFO] ')
    assert result.stdout.endswith('Hello, World, from cli.log.info!\nℹ Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')


def test_hello_no_color_verbose():
    result = check_command('./hello', '--no-color', '-v')
    check_returncode(result)
    assert result.stdout == '☐ You used -v you lucky person!\nℹ Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n'
